defmodule Tesla.Adapter.Httpc do
  @moduledoc """
    Adapter for `:httpc`

    **NOTE** Tesla overrides default autoredirect value with false to ensure
    consistency between adapters
  """

  import Tesla.Adapter.Shared, only: [stream_to_fun: 1, next_chunk: 1]

  @override_defaults autoredirect: false
  @http_opts ~w(timeout connect_timeout ssl essl autoredirect proxy_auth version relaxed url_encode)a

  def call(env, opts) do
    opts = Keyword.merge(@override_defaults, opts || [])
    with {:ok, {status, headers, body}} <- request(env, opts) do
      format_response(env, status, headers, body)
    end
  end

  defp format_response(env, {_, status, _}, headers, body) do
    %{env | status:   status,
            headers:  headers,
            body:     body}
  end

  defp request(env, opts) do
    content_type = to_char_list(env.headers["content-type"] || "")
    handle request(
      env.method || :get,
      Tesla.build_url(env.url, env.query) |> to_char_list,
      Enum.into(env.headers, [], fn {k,v} -> {to_char_list(k), to_char_list(v)} end),
      content_type,
      env.body,
      Keyword.split(opts ++ env.opts, @http_opts)
    )
  end

  defp request(method, url, headers, _content_type, nil, {http_opts, opts}) do
    :httpc.request(method, {url, headers}, http_opts, opts)
  end

  defp request(method, url, headers, content_type, %Stream{} = body, opts) do
    fun = stream_to_fun(body)
    request(method, url, headers, content_type, fun, opts)
  end

  defp request(method, url, headers, content_type, body, opts) when is_function(body) do
    body = {:chunkify, &next_chunk/1, body}
    request(method, url, headers, content_type, body, opts)
  end

  defp request(method, url, headers, content_type, body, {http_opts, opts}) do
    :httpc.request(method, {url, headers, content_type, body}, http_opts, opts)
  end

  defp handle({:error, {:failed_connect, _}}), do: {:error, :econnrefused}
  defp handle(response), do: response
end
